package main

import (
	"archive/zip"
	"bytes"
	"fmt"
	"html/template"
	"io"
	"io/ioutil"
	"net/http"
	"os"
	"path/filepath"
	"strconv"
	"strings"
)

type Data struct {
	ServiceNamePill, ServiceNameTitle, ServiceNameLower, Route, Port string
}

var Usage string

func init() {
	Usage = "Usage:\n  bee-bootstrap [<github-url> | node | python | go]"
}

func main() {
	if len(os.Args) < 2 {
		fmt.Println(Usage)
		os.Exit(0)
	}

	baseURL := os.Args[1]
	var fileURL string
	if strings.HasPrefix(baseURL, "https://") || strings.HasPrefix(baseURL, "http://") {
		fileURL = baseURL + "/archive/master.zip"
	} else if strings.HasPrefix(baseURL, "github.com/") {
		fileURL = "https://" + baseURL + "/archive/master.zip"
	} else {
		fileURL = "https://github.com/bee-travels/" + baseURL + "-service-template/archive/master.zip"
	}

	/*Service name (destination-basic):
	Service route (destinations):
	Service port (9201): */
	var serviceName string
	var route string
	var port string

	fmt.Print("Service name (destination-basic): ")
	fmt.Scan(&serviceName)

	fmt.Print("Service route (destinations): ")
	fmt.Scan(&route)

	fmt.Print("Service port (9000): ")
	fmt.Scan(&port)

	data, err := GetData(serviceName, route, port)
	if err != nil {
		fmt.Println("error : ", err)
		os.Exit(1)
	}

	err = DownloadFile(fileURL, "git.zip")
	if err != nil {
		panic(err)
	}

	err = Unzip("git.zip", "template")
	if err != nil {
		panic(err)
	}
	// ServiceNamePill, ServiceNameTitle, ServiceNameLower, Route, Port string
	Cleanup(data.ServiceNamePill)

	err = ProcessFiles("template", *data)
	if err != nil {
		panic(err)
	}

	pwd, err := os.Getwd()
	if err != nil {
		fmt.Println(err)
		panic(err)
	}

	err = MoveFile("./template", pwd+"/"+data.ServiceNamePill)
	if err != nil {
		panic(err)
	}
	fmt.Println("Done")
	Cleanup("./git.zip", "./template")
}

func GetData(serviceNamePill, route, port string) (*Data, error) {
	if serviceNamePill == "" || route == "" || port == "" {
		return nil, fmt.Errorf("no data provided")
	}

	if strings.Contains(serviceNamePill, " ") {
		return nil, fmt.Errorf("service name should not contain spaces")
	}

	if strings.Contains(serviceNamePill, "_") {
		return nil, fmt.Errorf("do not use _ in service names")
	}

	route = strings.ToLower(route)

	if _, err := strconv.Atoi(port); err != nil {
		return nil, fmt.Errorf("port should be a number")
	}

	serviceNamePill = strings.ToLower(serviceNamePill)
	serviceNameLower := strings.ReplaceAll(serviceNamePill, "-", " ")
	serviceNameTitle := strings.Title(serviceNameLower)

	return &Data{
		ServiceNamePill:  serviceNamePill,
		ServiceNameLower: serviceNameLower,
		ServiceNameTitle: serviceNameTitle,
		Route:            route,
		Port:             port,
	}, nil

}

func Cleanup(filepaths ...string) error {
	for _, filepath := range filepaths {
		fi, err := os.Stat(filepath)
		if err != nil {
			return err
		}
		mode := fi.Mode()
		if mode.IsDir() {
			if err := os.RemoveAll(filepath); err != nil {
				return err
			}
		} else if mode.IsRegular() {
			if err := os.Remove(filepath); err != nil {
				return err
			}
		}
	}
	return nil
}

func MoveFile(src, dst string) error {
	var files []string
	f, err := os.Open(src)
	if err != nil {
		return err
	}
	fileInfos, err := f.Readdir(-1)
	f.Close()
	if err != nil {
		return err
	}

	for _, file := range fileInfos {
		files = append(files, file.Name())
	}

	if len(files) != 1 {
		return fmt.Errorf("expected one file only")
	}

	base := src + "/" + files[0]
	if err := os.Rename(base, dst); err != nil {
		return err
	}
	return nil
}

func ProcessFiles(folder string, data Data) error {
	err := filepath.Walk(folder, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		if info.Mode().IsRegular() {
			fmt.Println(path)
			b, err := ioutil.ReadFile(path)
			if err != nil {
				return err
			}
			filecontent := string(b)
			fileTemplate := template.Must(template.New("file").Parse(filecontent))

			buf := new(bytes.Buffer)

			if err := fileTemplate.Execute(buf, data); err != nil {
				return err
			}

			if err := ioutil.WriteFile(path, buf.Bytes(), 0644); err != nil {
				return err
			}

			buf.Reset()

			filenameTemplate := template.Must(template.New("filename").Parse(info.Name()))
			if err := filenameTemplate.Execute(buf, data); err != nil {
				return err
			}

			filename := string(buf.Bytes())

			if filename != info.Name() {
				newPath := strings.ReplaceAll(path, info.Name(), filename)
				if err := os.Rename(path, newPath); err != nil {
					return err
				}
			}
		}
		return nil
	})
	return err
}

func Unzip(src, dest string) error {
	r, err := zip.OpenReader(src)
	if err != nil {
		return err
	}
	defer func() {
		if err := r.Close(); err != nil {
			panic(err)
		}
	}()

	os.MkdirAll(dest, 0755)

	// Closure to address file descriptors issue with all the deferred .Close() methods
	extractAndWriteFile := func(f *zip.File) error {
		rc, err := f.Open()
		if err != nil {
			return err
		}
		defer func() {
			if err := rc.Close(); err != nil {
				panic(err)
			}
		}()

		path := filepath.Join(dest, f.Name)

		if f.FileInfo().IsDir() {
			os.MkdirAll(path, f.Mode())
		} else {
			os.MkdirAll(filepath.Dir(path), f.Mode())
			f, err := os.OpenFile(path, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, f.Mode())
			if err != nil {
				return err
			}
			defer func() {
				if err := f.Close(); err != nil {
					panic(err)
				}
			}()

			_, err = io.Copy(f, rc)
			if err != nil {
				return err
			}
		}
		return nil
	}

	for _, f := range r.File {
		err := extractAndWriteFile(f)
		if err != nil {
			return err
		}
	}

	return nil
}

// DownloadFile ...
func DownloadFile(url, filepath string) error {
	resp, err := http.Get(url)
	if err != nil {
		return err
	}

	defer resp.Body.Close()

	out, err := os.Create(filepath)
	if err != nil {
		return err
	}

	defer out.Close()

	_, err = io.Copy(out, resp.Body)
	if err != nil {
		return err
	}
	return nil
}
